@@ -67,7 +67,7 @@ var externalCacheFormat = formatGzippedJSON{}
 
 // A SourceState is a source state.
 type SourceState struct {
-	entries                 map[RelPath]SourceStateEntry
+	root                    sourceStateEntryTreeNode
 	baseSystem              System
 	system                  System
 	sourceDirAbsPath        AbsPath
@@ -199,7 +199,6 @@ type targetStateEntryFunc func(System, AbsPath) (TargetStateEntry, error)
 // NewSourceState creates a new source state with the given options.
 func NewSourceState(options ...SourceStateOption) *SourceState {
 	s := &SourceState{
-		entries:              make(map[RelPath]SourceStateEntry),
 		umask:                Umask,
 		encryption:           NoEncryption{},
 		ignore:               newPatternSet(),
@@ -242,13 +241,11 @@ func (s *SourceState) Add(sourceSystem System, persistentState PersistentState,
 		sourceRelPaths []SourceRelPath
 	}
 
-	destAbsPaths := make([]AbsPath, 0, len(destAbsPathInfos))
+	destAbsPaths := make(AbsPaths, 0, len(destAbsPathInfos))
 	for destAbsPath := range destAbsPathInfos {
 		destAbsPaths = append(destAbsPaths, destAbsPath)
 	}
-	sort.Slice(destAbsPaths, func(i, j int) bool {
-		return destAbsPaths[i].absPath < destAbsPaths[j].absPath
-	})
+	sort.Sort(destAbsPaths)
 
 	sourceUpdates := make([]sourceUpdate, 0, len(destAbsPathInfos))
 	newSourceStateEntries := make(map[SourceRelPath]SourceStateEntry)
@@ -264,11 +261,11 @@ DESTABSPATH:
 
 		// Find the target's parent directory in the source state.
 		var parentSourceRelPath SourceRelPath
-		if targetParentRelPath := targetRelPath.Dir(); targetParentRelPath == "." {
+		if targetParentRelPath := targetRelPath.Dir(); targetParentRelPath == DotRelPath {
 			parentSourceRelPath = SourceRelPath{}
 		} else if parentEntry, ok := newSourceStateEntriesByTargetRelPath[targetParentRelPath]; ok {
 			parentSourceRelPath = parentEntry.SourceRelPath()
-		} else if parentEntry, ok := s.entries[targetParentRelPath]; ok {
+		} else if parentEntry := s.root.Get(targetParentRelPath); parentEntry != nil {
 			parentSourceRelPath = parentEntry.SourceRelPath()
 		} else {
 			return fmt.Errorf("%s: parent directory not in source state", destAbsPath)
@@ -299,7 +296,7 @@ DESTABSPATH:
 			sourceRelPaths: []SourceRelPath{sourceEntryRelPath},
 		}
 
-		if oldSourceStateEntry, ok := s.entries[targetRelPath]; ok {
+		if oldSourceStateEntry := s.root.Get(targetRelPath); oldSourceStateEntry != nil {
 			oldSourceEntryRelPath := oldSourceStateEntry.SourceRelPath()
 			if !oldSourceEntryRelPath.Empty() && oldSourceEntryRelPath != sourceEntryRelPath {
 				if options.PreAddFunc != nil {
@@ -362,24 +359,24 @@ DESTABSPATH:
 		}
 	}
 
-	sourceEntries := make(map[RelPath]SourceStateEntry)
+	var sourceRoot sourceStateEntryTreeNode
 	for sourceRelPath, sourceStateEntry := range newSourceStateEntries {
-		sourceEntries[sourceRelPath.RelPath()] = sourceStateEntry
+		sourceRoot.Set(sourceRelPath.RelPath(), sourceStateEntry)
 	}
 
 	// Simulate removing a directory by creating SourceStateRemove entries for
 	// all existing source state entries that are in options.RemoveDir and not
 	// in the new source state.
-	if options.RemoveDir != "" {
-		for targetRelPath, sourceStateEntry := range s.entries {
+	if options.RemoveDir != EmptyRelPath {
+		_ = s.root.ForEach(EmptyRelPath, func(targetRelPath RelPath, sourceStateEntry SourceStateEntry) error {
 			if !targetRelPath.HasDirPrefix(options.RemoveDir) {
-				continue
+				return nil
 			}
 			if _, ok := newSourceStateEntriesByTargetRelPath[targetRelPath]; ok {
-				continue
+				return nil
 			}
 			sourceRelPath := sourceStateEntry.SourceRelPath()
-			sourceEntries[sourceRelPath.RelPath()] = &SourceStateRemove{}
+			sourceRoot.Set(sourceRelPath.RelPath(), &SourceStateRemove{})
 			update := sourceUpdate{
 				destAbsPath: s.destDirAbsPath.Join(targetRelPath),
 				entryState: &EntryState{
@@ -388,11 +385,12 @@ DESTABSPATH:
 				sourceRelPaths: []SourceRelPath{sourceRelPath},
 			}
 			sourceUpdates = append(sourceUpdates, update)
-		}
+			return nil
+		})
 	}
 
 	targetSourceState := &SourceState{
-		entries: sourceEntries,
+		root: sourceRoot,
 	}
 
 	for _, sourceUpdate := range sourceUpdates {
@@ -440,7 +438,7 @@ func (s *SourceState) AddDestAbsPathInfos(destAbsPathInfos map[AbsPath]fs.FileIn
 			return nil
 		}
 		parentRelPath := parentAbsPath.MustTrimDirPrefix(s.destDirAbsPath)
-		if _, ok := s.entries[parentRelPath]; ok {
+		if s.root.Get(parentRelPath) != nil {
 			return nil
 		}
 
@@ -461,7 +459,7 @@ type ApplyOptions struct {
 
 // Apply updates targetRelPath in targetDir in destSystem to match s.
 func (s *SourceState) Apply(targetSystem, destSystem System, persistentState PersistentState, targetDir AbsPath, targetRelPath RelPath, options ApplyOptions) error {
-	sourceStateEntry := s.entries[targetRelPath]
+	sourceStateEntry := s.root.Get(targetRelPath)
 
 	if !options.Include.IncludeEncrypted() {
 		if sourceStateFile, ok := sourceStateEntry.(*SourceStateFile); ok && sourceStateFile.Attr.Encrypted {
@@ -595,22 +593,16 @@ func (s *SourceState) Apply(targetSystem, destSystem System, persistentState Per
 	return persistentStateSet(persistentState, EntryStateBucket, targetAbsPath.Bytes(), targetEntryState)
 }
 
+// Contains returns the source state entry for targetRelPath.
+func (s *SourceState) Contains(targetRelPath RelPath) bool {
+	return s.root.Get(targetRelPath) != nil
+}
+
 // Encryption returns s's encryption.
 func (s *SourceState) Encryption() Encryption {
 	return s.encryption
 }
 
-// Entries returns s's source state entries.
-func (s *SourceState) Entries() map[RelPath]SourceStateEntry {
-	return s.entries
-}
-
-// Entry returns the source state entry for targetRelPath.
-func (s *SourceState) Entry(targetRelPath RelPath) (SourceStateEntry, bool) {
-	sourceStateEntry, ok := s.entries[targetRelPath]
-	return sourceStateEntry, ok
-}
-
 // ExecuteTemplateData returns the result of executing template data.
 func (s *SourceState) ExecuteTemplateData(name string, data []byte) ([]byte, error) {
 	tmpl, err := template.New(name).
@@ -642,6 +634,13 @@ func (s *SourceState) ExecuteTemplateData(name string, data []byte) ([]byte, err
 	return []byte(builder.String()), nil
 }
 
+// ForEach calls f for each source state entry.
+func (s *SourceState) ForEach(f func(RelPath, SourceStateEntry) error) error {
+	return s.root.ForEach(EmptyRelPath, func(targetRelPath RelPath, entry SourceStateEntry) error {
+		return f(targetRelPath, entry)
+	})
+}
+
 // Ignore returns if targetRelPath should be ignored.
 func (s *SourceState) Ignore(targetRelPath RelPath) bool {
 	return s.ignore.match(targetRelPath.String())
@@ -655,8 +654,8 @@ func (s *SourceState) MinVersion() semver.Version {
 // MustEntry returns the source state entry associated with targetRelPath, and
 // panics if it does not exist.
 func (s *SourceState) MustEntry(targetRelPath RelPath) SourceStateEntry {
-	sourceStateEntry, ok := s.entries[targetRelPath]
-	if !ok {
+	sourceStateEntry := s.root.Get(targetRelPath)
+	if sourceStateEntry == nil {
 		panic(fmt.Sprintf("%s: not in source state", targetRelPath))
 	}
 	return sourceStateEntry
@@ -736,7 +735,7 @@ func (s *SourceState) Read(ctx context.Context, options *ReadOptions) error {
 			}
 			n := 0
 			for _, match := range matches {
-				if !s.Ignore(RelPath(match)) {
+				if !s.Ignore(NewRelPath(match)) {
 					matches[n] = match
 					n++
 				}
@@ -744,7 +743,7 @@ func (s *SourceState) Read(ctx context.Context, options *ReadOptions) error {
 			targetParentRelPath := parentSourceRelPath.TargetRelPath(s.encryption.EncryptedSuffix())
 			matches = matches[:n]
 			for _, match := range matches {
-				targetRelPath := targetParentRelPath.JoinStr(match)
+				targetRelPath := targetParentRelPath.JoinString(match)
 				sourceStateEntry := &SourceStateRemove{
 					targetRelPath: targetRelPath,
 				}
@@ -767,7 +766,7 @@ func (s *SourceState) Read(ctx context.Context, options *ReadOptions) error {
 			return nil
 		case info.IsDir():
 			da := parseDirAttr(sourceName.String())
-			targetRelPath := parentSourceRelPath.Dir().TargetRelPath(s.encryption.EncryptedSuffix()).Join(RelPath(da.TargetName))
+			targetRelPath := parentSourceRelPath.Dir().TargetRelPath(s.encryption.EncryptedSuffix()).JoinString(da.TargetName)
 			if s.Ignore(targetRelPath) {
 				return vfs.SkipDir
 			}
@@ -776,7 +775,7 @@ func (s *SourceState) Read(ctx context.Context, options *ReadOptions) error {
 			return nil
 		case info.Mode().IsRegular():
 			fa := parseFileAttr(sourceName.String(), s.encryption.EncryptedSuffix())
-			targetRelPath := parentSourceRelPath.Dir().TargetRelPath(s.encryption.EncryptedSuffix()).Join(RelPath(fa.TargetName))
+			targetRelPath := parentSourceRelPath.Dir().TargetRelPath(s.encryption.EncryptedSuffix()).JoinString(fa.TargetName)
 			if s.Ignore(targetRelPath) {
 				return nil
 			}
@@ -795,18 +794,25 @@ func (s *SourceState) Read(ctx context.Context, options *ReadOptions) error {
 	}
 
 	// Read externals.
-	externalRelPaths := make([]RelPath, 0, len(s.externals))
+	externalRelPaths := make(RelPaths, 0, len(s.externals))
 	for externalRelPath := range s.externals {
 		externalRelPaths = append(externalRelPaths, externalRelPath)
 	}
-	sort.Slice(externalRelPaths, func(i, j int) bool {
-		return externalRelPaths[i] < externalRelPaths[j]
-	})
+	sort.Sort(externalRelPaths)
 	for _, externalRelPath := range externalRelPaths {
 		if s.Ignore(externalRelPath) {
 			continue
 		}
-		externalSourceStateEntries, err := s.readExternal(ctx, externalRelPath, s.externals[externalRelPath], options)
+		external := s.externals[externalRelPath]
+		parentRelPath, _ := externalRelPath.Split()
+		var parentSourceRelPath SourceRelPath
+		switch parentSourceStateEntry, err := s.root.MkdirAll(parentRelPath, external.URL, s.umask); {
+		case err != nil:
+			return err
+		case parentSourceStateEntry != nil:
+			parentSourceRelPath = parentSourceStateEntry.SourceRelPath()
+		}
+		externalSourceStateEntries, err := s.readExternal(ctx, externalRelPath, parentSourceRelPath, external, options)
 		if err != nil {
 			return err
 		}
@@ -845,7 +851,7 @@ func (s *SourceState) Read(ctx context.Context, options *ReadOptions) error {
 				if name == "." || name == ".." {
 					continue
 				}
-				destEntryRelPath := targetRelPath.JoinStr(name)
+				destEntryRelPath := targetRelPath.JoinString(name)
 				if _, ok := allSourceStateEntries[destEntryRelPath]; ok {
 					continue
 				}
@@ -863,31 +869,33 @@ func (s *SourceState) Read(ctx context.Context, options *ReadOptions) error {
 		}
 	}
 
-	// Check for duplicate source entries with the same target name. Iterate
-	// over the target names in order so that any error is deterministic.
-	targetRelPaths := make([]RelPath, 0, len(allSourceStateEntries))
+	// Check for inconsistent source entries. Iterate over the target names in
+	// order so that any error is deterministic.
+	targetRelPaths := make(RelPaths, 0, len(allSourceStateEntries))
 	for targetRelPath := range allSourceStateEntries {
 		targetRelPaths = append(targetRelPaths, targetRelPath)
 	}
-	sort.Slice(targetRelPaths, func(i, j int) bool {
-		return targetRelPaths[i] < targetRelPaths[j]
-	})
+	sort.Sort(targetRelPaths)
 	var err error
 	for _, targetRelPath := range targetRelPaths {
 		sourceStateEntries := allSourceStateEntries[targetRelPath]
 		if len(sourceStateEntries) == 1 {
 			continue
 		}
-		sourceRelPaths := make([]SourceRelPath, 0, len(sourceStateEntries))
-		for _, sourceStateEntry := range sourceStateEntries {
-			sourceRelPaths = append(sourceRelPaths, sourceStateEntry.SourceRelPath())
+
+		// Allow duplicate equivalent source entries for directories.
+		if allEquivalentDirs(sourceStateEntries) {
+			continue
 		}
-		sort.Slice(sourceRelPaths, func(i, j int) bool {
-			return sourceRelPaths[i].relPath < sourceRelPaths[j].relPath
-		})
-		err = multierr.Append(err, &duplicateTargetError{
-			targetRelPath:  targetRelPath,
-			sourceRelPaths: sourceRelPaths,
+
+		origins := make([]string, 0, len(sourceStateEntries))
+		for _, sourceStateEntry := range sourceStateEntries {
+			origins = append(origins, sourceStateEntry.Origin())
+		}
+		sort.Strings(origins)
+		err = multierr.Append(err, &inconsistentStateError{
+			targetRelPath: targetRelPath,
+			origins:       origins,
 		})
 	}
 	if err != nil {
@@ -896,7 +904,7 @@ func (s *SourceState) Read(ctx context.Context, options *ReadOptions) error {
 
 	// Populate s.Entries with the unique source entry for each target.
 	for targetRelPath, sourceEntries := range allSourceStateEntries {
-		s.entries[targetRelPath] = sourceEntries[0]
+		s.root.Set(targetRelPath, sourceEntries[0])
 	}
 
 	return nil
@@ -904,18 +912,19 @@ func (s *SourceState) Read(ctx context.Context, options *ReadOptions) error {
 
 // TargetRelPaths returns all of s's target relative paths in order.
 func (s *SourceState) TargetRelPaths() []RelPath {
-	targetRelPaths := make([]RelPath, 0, len(s.entries))
-	for targetRelPath := range s.entries {
+	entries := s.root.Map()
+	targetRelPaths := make([]RelPath, 0, len(entries))
+	for targetRelPath := range entries {
 		targetRelPaths = append(targetRelPaths, targetRelPath)
 	}
 	sort.Slice(targetRelPaths, func(i, j int) bool {
-		orderI := s.entries[targetRelPaths[i]].Order()
-		orderJ := s.entries[targetRelPaths[j]].Order()
+		orderI := entries[targetRelPaths[i]].Order()
+		orderJ := entries[targetRelPaths[j]].Order()
 		switch {
 		case orderI < orderJ:
 			return true
 		case orderI == orderJ:
-			return targetRelPaths[i] < targetRelPaths[j]
+			return targetRelPaths[i].Less(targetRelPaths[j])
 		default:
 			return false
 		}
@@ -945,7 +954,7 @@ func (s *SourceState) addExternal(sourceAbsPath AbsPath) error {
 	if err != nil {
 		return err
 	}
-	parentSourceRelPath := NewSourceRelDirPath(parentRelPath)
+	parentSourceRelPath := NewSourceRelDirPath(parentRelPath.String())
 	parentTargetSourceRelPath := parentSourceRelPath.TargetRelPath(s.encryption.EncryptedSuffix())
 
 	format, ok := Formats[strings.TrimPrefix(sourceAbsPath.Ext(), ".")]
@@ -956,12 +965,12 @@ func (s *SourceState) addExternal(sourceAbsPath AbsPath) error {
 	if err != nil {
 		return fmt.Errorf("%s: %w", sourceAbsPath, err)
 	}
-	externals := make(map[RelPath]External)
+	externals := make(map[string]External)
 	if err := format.Unmarshal(data, &externals); err != nil {
 		return fmt.Errorf("%s: %w", sourceAbsPath, err)
 	}
-	for relPath, external := range externals {
-		targetRelPath := parentTargetSourceRelPath.Join(relPath)
+	for relPathStr, external := range externals {
+		targetRelPath := parentTargetSourceRelPath.JoinString(relPathStr)
 		if _, ok := s.externals[targetRelPath]; ok {
 			return fmt.Errorf("%s: duplicate externals", targetRelPath)
 		}
@@ -995,7 +1004,7 @@ func (s *SourceState) addPatterns(patternSet *patternSet, sourceAbsPath AbsPath,
 			include = false
 			text = mustTrimPrefix(text, "!")
 		}
-		pattern := string(dir.JoinStr(text))
+		pattern := dir.JoinString(text).String()
 		if err := patternSet.add(pattern, include); err != nil {
 			return fmt.Errorf("%s:%d: %w", sourceAbsPath, lineNumber, err)
 		}
@@ -1036,7 +1045,7 @@ func (s *SourceState) addTemplatesDir(templatesDirAbsPath AbsPath) error {
 				return err
 			}
 			templateRelPath := templateAbsPath.MustTrimDirPrefix(templatesDirAbsPath)
-			name := string(templateRelPath)
+			name := templateRelPath.String()
 			tmpl, err := template.New(name).Option(s.templateOptions...).Funcs(s.templateFuncs).Parse(string(contents))
 			if err != nil {
 				return err
@@ -1099,7 +1108,7 @@ func (s *SourceState) getExternalDataRaw(ctx context.Context, externalRelPath Re
 	now = now.UTC()
 
 	cacheKey := hex.EncodeToString(SHA256Sum([]byte(external.URL)))
-	cachedDataAbsPath := s.cacheDirAbsPath.JoinStr("external", cacheKey+"."+externalCacheFormat.Name())
+	cachedDataAbsPath := s.cacheDirAbsPath.JoinString("external", cacheKey+"."+externalCacheFormat.Name())
 	if options == nil || !options.RefreshExternals {
 		if data, err := s.system.ReadFile(cachedDataAbsPath); err == nil {
 			var externalCacheEntry externalCacheEntry
@@ -1188,6 +1197,7 @@ func (s *SourceState) newSourceStateDir(sourceRelPath SourceRelPath, dirAttr Dir
 		perm: dirAttr.perm() &^ s.umask,
 	}
 	return &SourceStateDir{
+		origin:           sourceRelPath.String(),
 		sourceRelPath:    sourceRelPath,
 		Attr:             dirAttr,
 		targetStateEntry: targetStateDir,
@@ -1299,8 +1309,7 @@ func (s *SourceState) newModifyTargetStateEntryFunc(sourceRelPath SourceRelPath,
 
 			// Write the modifier to a temporary file.
 			var tempFile *os.File
-			tempFile, err = os.CreateTemp("", "*."+fileAttr.TargetName)
-			if err != nil {
+			if tempFile, err = os.CreateTemp("", "*."+fileAttr.TargetName); err != nil {
 				return
 			}
 			defer func() {
@@ -1321,7 +1330,8 @@ func (s *SourceState) newModifyTargetStateEntryFunc(sourceRelPath SourceRelPath,
 			cmd := interpreter.ExecCommand(tempFile.Name())
 			cmd.Stdin = bytes.NewReader(currentContents)
 			cmd.Stderr = os.Stderr
-			return destSystem.IdempotentCmdOutput(cmd)
+			contents, err = destSystem.IdempotentCmdOutput(cmd)
+			return
 		}
 		return &TargetStateFile{
 			lazyContents: newLazyContentsFunc(contentsFunc),
@@ -1420,7 +1430,7 @@ func (s *SourceState) newSourceStateFile(sourceRelPath SourceRelPath, fileAttr F
 		if interpreter != nil {
 			// For modify scripts, the script extension is not considered part
 			// of the target name, so remove it.
-			targetRelPath = targetRelPath[:len(targetRelPath)-len(ext)-1]
+			targetRelPath = targetRelPath.Slice(0, targetRelPath.Len()-len(ext)-1)
 		}
 		targetStateEntryFunc = s.newModifyTargetStateEntryFunc(sourceRelPath, fileAttr, sourceLazyContents, interpreter)
 	case SourceFileTypeRemove:
@@ -1439,6 +1449,7 @@ func (s *SourceState) newSourceStateFile(sourceRelPath SourceRelPath, fileAttr F
 
 	return targetRelPath, &SourceStateFile{
 		lazyContents:         sourceLazyContents,
+		origin:               sourceRelPath.String(),
 		sourceRelPath:        sourceRelPath,
 		Attr:                 fileAttr,
 		targetStateEntryFunc: targetStateEntryFunc,
@@ -1457,9 +1468,10 @@ func (s *SourceState) newSourceStateDirEntry(info fs.FileInfo, parentSourceRelPa
 		Private:    isPrivate(info),
 		ReadOnly:   isReadOnly(info),
 	}
-	sourceRelPath := parentSourceRelPath.Join(NewSourceRelDirPath(RelPath(dirAttr.SourceName())))
+	sourceRelPath := parentSourceRelPath.Join(NewSourceRelDirPath(dirAttr.SourceName()))
 	return &SourceStateDir{
 		Attr:          dirAttr,
+		origin:        sourceRelPath.String(),
 		sourceRelPath: sourceRelPath,
 		targetStateEntry: &TargetStateDir{
 			perm: 0o777 &^ s.umask,
@@ -1508,9 +1520,10 @@ func (s *SourceState) newSourceStateFileEntryFromFile(actualStateFile *ActualSta
 		}
 	}
 	lazyContents := newLazyContents(contents)
-	sourceRelPath := parentSourceRelPath.Join(NewSourceRelPath(RelPath(fileAttr.SourceName(s.encryption.EncryptedSuffix()))))
+	sourceRelPath := parentSourceRelPath.Join(NewSourceRelPath(fileAttr.SourceName(s.encryption.EncryptedSuffix())))
 	return &SourceStateFile{
 		Attr:          fileAttr,
+		origin:        sourceRelPath.String(),
 		sourceRelPath: sourceRelPath,
 		lazyContents:  lazyContents,
 		targetStateEntry: &TargetStateFile{
@@ -1555,7 +1568,7 @@ func (s *SourceState) newSourceStateFileEntryFromSymlink(actualStateSymlink *Act
 		Type:       SourceFileTypeSymlink,
 		Template:   template,
 	}
-	sourceRelPath := parentSourceRelPath.Join(NewSourceRelPath(RelPath(fileAttr.SourceName(s.encryption.EncryptedSuffix()))))
+	sourceRelPath := parentSourceRelPath.Join(NewSourceRelPath(fileAttr.SourceName(s.encryption.EncryptedSuffix())))
 	return &SourceStateFile{
 		Attr:          fileAttr,
 		sourceRelPath: sourceRelPath,
@@ -1568,12 +1581,12 @@ func (s *SourceState) newSourceStateFileEntryFromSymlink(actualStateSymlink *Act
 }
 
 // readExternal reads an external and returns its SourceStateEntries.
-func (s *SourceState) readExternal(ctx context.Context, externalRelPath RelPath, external External, options *ReadOptions) (map[RelPath][]SourceStateEntry, error) {
+func (s *SourceState) readExternal(ctx context.Context, externalRelPath RelPath, parentSourceRelPath SourceRelPath, external External, options *ReadOptions) (map[RelPath][]SourceStateEntry, error) {
 	switch external.Type {
 	case ExternalTypeArchive:
-		return s.readExternalArchive(ctx, externalRelPath, external, options)
+		return s.readExternalArchive(ctx, externalRelPath, parentSourceRelPath, external, options)
 	case ExternalTypeFile:
-		return s.readExternalFile(ctx, externalRelPath, external, options)
+		return s.readExternalFile(ctx, externalRelPath, parentSourceRelPath, external, options)
 	default:
 		return nil, fmt.Errorf("%s: unknown external type: %s", externalRelPath, external.Type)
 	}
@@ -1581,7 +1594,7 @@ func (s *SourceState) readExternal(ctx context.Context, externalRelPath RelPath,
 
 // readExternalArchive reads an external archive and returns its
 // SourceStateEntries.
-func (s *SourceState) readExternalArchive(ctx context.Context, externalRelPath RelPath, external External, options *ReadOptions) (map[RelPath][]SourceStateEntry, error) {
+func (s *SourceState) readExternalArchive(ctx context.Context, externalRelPath RelPath, parentSourceRelPath SourceRelPath, external External, options *ReadOptions) (map[RelPath][]SourceStateEntry, error) {
 	data, err := s.getExternalData(ctx, externalRelPath, external, options)
 	if err != nil {
 		return nil, err
@@ -1595,27 +1608,28 @@ func (s *SourceState) readExternalArchive(ctx context.Context, externalRelPath R
 	if external.Encrypted {
 		urlPath = strings.TrimSuffix(urlPath, s.encryption.EncryptedSuffix())
 	}
-	sourceRelPath := NewSourceRelPath(RelPath(external.URL))
-	sourceStateEntries := map[RelPath][]SourceStateEntry{
-		externalRelPath: {
-			&SourceStateDir{
-				Attr: DirAttr{
-					TargetName: externalRelPath.Base(),
-					Exact:      external.Exact,
-				},
-				sourceRelPath: sourceRelPath,
-				targetStateEntry: &TargetStateDir{
-					perm: 0o777 &^ s.umask,
-				},
-			},
+	dirAttr := DirAttr{
+		TargetName: externalRelPath.Base(),
+		Exact:      external.Exact,
+	}
+	sourceStateDir := &SourceStateDir{
+		Attr:          dirAttr,
+		origin:        external.URL,
+		sourceRelPath: parentSourceRelPath.Join(NewSourceRelPath(dirAttr.SourceName())),
+		targetStateEntry: &TargetStateDir{
+			perm: 0o777 &^ s.umask,
 		},
 	}
+	sourceStateEntries := map[RelPath][]SourceStateEntry{
+		externalRelPath: {sourceStateDir},
+	}
 
 	format := external.Format
 	if format == ArchiveFormatUnknown {
 		format = GuessArchiveFormat(urlPath, data)
 	}
 
+	sourceRelPaths := make(map[RelPath]SourceRelPath)
 	if err := walkArchive(data, format, func(name string, info fs.FileInfo, r io.Reader, linkname string) error {
 		if external.StripComponents > 0 {
 			components := strings.Split(name, "/")
@@ -1627,26 +1641,31 @@ func (s *SourceState) readExternalArchive(ctx context.Context, externalRelPath R
 		if name == "" {
 			return nil
 		}
-		targetRelPath := externalRelPath.JoinStr(name)
+		targetRelPath := externalRelPath.JoinString(name)
 
 		if s.Ignore(targetRelPath) {
 			return nil
 		}
 
+		dirTargetRelPath, _ := targetRelPath.Split()
+		dirSourceRelPath := sourceRelPaths[dirTargetRelPath]
+
 		var sourceStateEntry SourceStateEntry
 		switch {
 		case info.IsDir():
 			targetStateEntry := &TargetStateDir{
 				perm: info.Mode().Perm() &^ s.umask,
 			}
+			dirAttr := DirAttr{
+				TargetName: info.Name(),
+				Exact:      external.Exact,
+				Private:    isPrivate(info),
+				ReadOnly:   isReadOnly(info),
+			}
 			sourceStateEntry = &SourceStateDir{
-				Attr: DirAttr{
-					TargetName: info.Name(),
-					Exact:      external.Exact,
-					Private:    isPrivate(info),
-					ReadOnly:   isReadOnly(info),
-				},
-				sourceRelPath:    sourceRelPath,
+				Attr:             dirAttr,
+				origin:           external.URL,
+				sourceRelPath:    parentSourceRelPath.Join(dirSourceRelPath, NewSourceRelPath(dirAttr.SourceName())),
 				targetStateEntry: targetStateEntry,
 			}
 		case info.Mode()&fs.ModeType == 0:
@@ -1671,19 +1690,22 @@ func (s *SourceState) readExternalArchive(ctx context.Context, externalRelPath R
 			sourceStateEntry = &SourceStateFile{
 				lazyContents:     lazyContents,
 				Attr:             fileAttr,
-				sourceRelPath:    sourceRelPath,
+				origin:           external.URL,
+				sourceRelPath:    parentSourceRelPath.Join(dirSourceRelPath, NewSourceRelPath(fileAttr.SourceName(s.encryption.EncryptedSuffix()))),
 				targetStateEntry: targetStateEntry,
 			}
 		case info.Mode()&fs.ModeType == fs.ModeSymlink:
 			targetStateEntry := &TargetStateSymlink{
 				lazyLinkname: newLazyLinkname(linkname),
 			}
+			fileAttr := FileAttr{
+				TargetName: info.Name(),
+				Type:       SourceFileTypeSymlink,
+			}
 			sourceStateEntry = &SourceStateFile{
-				Attr: FileAttr{
-					TargetName: info.Name(),
-					Type:       SourceFileTypeSymlink,
-				},
-				sourceRelPath:    sourceRelPath,
+				Attr:             fileAttr,
+				origin:           external.URL,
+				sourceRelPath:    parentSourceRelPath.Join(dirSourceRelPath, NewSourceRelPath(fileAttr.SourceName(s.encryption.EncryptedSuffix()))),
 				targetStateEntry: targetStateEntry,
 			}
 		default:
@@ -1699,7 +1721,7 @@ func (s *SourceState) readExternalArchive(ctx context.Context, externalRelPath R
 }
 
 // readExternalFile reads an external file and returns its SourceStateEntries.
-func (s *SourceState) readExternalFile(ctx context.Context, externalRelPath RelPath, external External, options *ReadOptions) (map[RelPath][]SourceStateEntry, error) {
+func (s *SourceState) readExternalFile(ctx context.Context, externalRelPath RelPath, parentSourceRelPath SourceRelPath, external External, options *ReadOptions) (map[RelPath][]SourceStateEntry, error) {
 	lazyContents := newLazyContentsFunc(func() ([]byte, error) {
 		return s.getExternalData(ctx, externalRelPath, external, options)
 	})
@@ -1713,7 +1735,8 @@ func (s *SourceState) readExternalFile(ctx context.Context, externalRelPath RelP
 		perm:         fileAttr.perm() &^ s.umask,
 	}
 	sourceStateEntry := &SourceStateFile{
-		sourceRelPath:    NewSourceRelPath(RelPath(external.URL)),
+		origin:           external.URL,
+		sourceRelPath:    parentSourceRelPath.Join(NewSourceRelPath(fileAttr.SourceName(s.encryption.EncryptedSuffix()))),
 		targetStateEntry: targetStateEntry,
 	}
 	return map[RelPath][]SourceStateEntry{
@@ -1736,3 +1759,22 @@ func (s *SourceState) sourceStateEntry(actualStateEntry ActualStateEntry, destAb
 		panic(fmt.Sprintf("%T: unsupported type", actualStateEntry))
 	}
 }
+
+// allEquivalentDirs returns if sourceStateEntries are all equivalent
+// directories.
+func allEquivalentDirs(sourceStateEntries []SourceStateEntry) bool {
+	sourceStateDir0, ok := sourceStateEntries[0].(*SourceStateDir)
+	if !ok {
+		return false
+	}
+	for _, sourceStateEntry := range sourceStateEntries[1:] {
+		sourceStateDir, ok := sourceStateEntry.(*SourceStateDir)
+		if !ok {
+			return false
+		}
+		if sourceStateDir0.Attr != sourceStateDir.Attr {
+			return false
+		}
+	}
+	return true
+}
