@@ -547,6 +547,72 @@ func TestSourceStateAdd(t *testing.T) {
 	}
 }
 
+func TestSourceStateAddInExternal(t *testing.T) {
+	buffer := &bytes.Buffer{}
+	tarWriterSystem := NewTARWriterSystem(buffer, tar.Header{})
+	require.NoError(t, tarWriterSystem.Mkdir(NewAbsPath("dir"), 0o777))
+	require.NoError(t, tarWriterSystem.WriteFile(NewAbsPath("dir/file"), []byte("# contents of dir/file\n"), 0o666))
+	require.NoError(t, tarWriterSystem.Close())
+	archiveData := buffer.Bytes()
+
+	httpServer := httptest.NewServer(http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
+		_, err := w.Write(archiveData)
+		require.NoError(t, err)
+	}))
+	defer httpServer.Close()
+
+	root := map[string]interface{}{
+		"/home/user": map[string]interface{}{
+			".dir/file2": "# contents of .dir/file2\n",
+			".local/share/chezmoi": map[string]interface{}{
+				".chezmoiexternal.toml": chezmoitest.JoinLines(
+					`[".dir"]`,
+					`    type = "archive"`,
+					`    url = "`+httpServer.URL+`/archive.tar"`,
+					`    stripComponents = 1`,
+				),
+				"dot_dir": &vfst.Dir{Perm: 0o777},
+			},
+		},
+	}
+
+	chezmoitest.WithTestFS(t, root, func(fileSystem vfs.FS) {
+		ctx := context.Background()
+		system := NewRealSystem(fileSystem)
+		persistentState := NewMockPersistentState()
+		s := NewSourceState(
+			WithBaseSystem(system),
+			WithCacheDir(NewAbsPath("/home/user/.cache/chezmoi")),
+			WithDestDir(NewAbsPath("/home/user")),
+			WithSourceDir(NewAbsPath("/home/user/.local/share/chezmoi")),
+			WithSystem(system),
+		)
+		require.NoError(t, s.Read(ctx, nil))
+
+		destAbsPath := NewAbsPath("/home/user/.dir/file2")
+		fileInfo, err := system.Stat(destAbsPath)
+		require.NoError(t, err)
+		destAbsPathInfos := map[AbsPath]fs.FileInfo{
+			destAbsPath: fileInfo,
+		}
+		require.NoError(t, s.Add(system, persistentState, system, destAbsPathInfos, &AddOptions{
+			Include: NewEntryTypeSet(EntryTypesAll),
+		}))
+
+		vfst.RunTests(t, fileSystem, "",
+			vfst.TestPath("/home/user/.local/share/chezmoi/dot_dir",
+				vfst.TestIsDir,
+				vfst.TestModePerm(0o777&^chezmoitest.Umask),
+			),
+			vfst.TestPath("/home/user/.local/share/chezmoi/dot_dir/file2",
+				vfst.TestModeIsRegular,
+				vfst.TestModePerm(0o666&^chezmoitest.Umask),
+				vfst.TestContentsString("# contents of .dir/file2\n"),
+			),
+		)
+	})
+}
+
 func TestSourceStateApplyAll(t *testing.T) {
 	for _, tc := range []struct {
 		name               string
@@ -791,7 +857,8 @@ func TestSourceStateRead(t *testing.T) {
 			},
 			expectedSourceState: NewSourceState(
 				withEntries(map[RelPath]SourceStateEntry{
-					"dir": &SourceStateDir{
+					NewRelPath("dir"): &SourceStateDir{
+						origin:        "dir",
 						sourceRelPath: NewSourceRelDirPath("dir"),
 						Attr: DirAttr{
 							TargetName: "dir",
@@ -812,7 +879,8 @@ func TestSourceStateRead(t *testing.T) {
 			},
 			expectedSourceState: NewSourceState(
 				withEntries(map[RelPath]SourceStateEntry{
-					".file": &SourceStateFile{
+					NewRelPath(".file"): &SourceStateFile{
+						origin:        "dot_file",
 						sourceRelPath: NewSourceRelPath("dot_file"),
 						Attr: FileAttr{
 							TargetName: ".file",
@@ -835,7 +903,7 @@ func TestSourceStateRead(t *testing.T) {
 					"dot_file.tmpl": "# contents of .file\n",
 				},
 			},
-			expectedError: ".file: duplicate source state entries (dot_file, dot_file.tmpl)",
+			expectedError: ".file: inconsistent state (dot_file, dot_file.tmpl)",
 		},
 		{
 			name: "duplicate_target_dir",
@@ -849,7 +917,7 @@ func TestSourceStateRead(t *testing.T) {
 					},
 				},
 			},
-			expectedError: "dir: duplicate source state entries (dir, exact_dir)",
+			expectedError: "dir: inconsistent state (dir, exact_dir)",
 		},
 		{
 			name: "duplicate_target_script",
@@ -859,7 +927,7 @@ func TestSourceStateRead(t *testing.T) {
 					"run_once_script": "#!/bin/sh\n",
 				},
 			},
-			expectedError: "script: duplicate source state entries (run_once_script, run_script)",
+			expectedError: "script: inconsistent state (run_once_script, run_script)",
 		},
 		{
 			name: "symlink_with_attr",
@@ -871,7 +939,8 @@ func TestSourceStateRead(t *testing.T) {
 			},
 			expectedSourceState: NewSourceState(
 				withEntries(map[RelPath]SourceStateEntry{
-					".file": &SourceStateFile{
+					NewRelPath(".file"): &SourceStateFile{
+						origin:        "executable_dot_file",
 						sourceRelPath: NewSourceRelPath("executable_dot_file"),
 						Attr: FileAttr{
 							TargetName: ".file",
@@ -897,7 +966,8 @@ func TestSourceStateRead(t *testing.T) {
 			},
 			expectedSourceState: NewSourceState(
 				withEntries(map[RelPath]SourceStateEntry{
-					"script": &SourceStateFile{
+					NewRelPath("script"): &SourceStateFile{
+						origin:        "run_script",
 						sourceRelPath: NewSourceRelPath("run_script"),
 						Attr: FileAttr{
 							TargetName: "script",
@@ -905,7 +975,7 @@ func TestSourceStateRead(t *testing.T) {
 						},
 						lazyContents: newLazyContents([]byte("# contents of .script\n")),
 						targetStateEntry: &TargetStateScript{
-							name:         "script",
+							name:         NewRelPath("script"),
 							lazyContents: newLazyContents([]byte("# contents of .script\n")),
 						},
 					},
@@ -921,7 +991,8 @@ func TestSourceStateRead(t *testing.T) {
 			},
 			expectedSourceState: NewSourceState(
 				withEntries(map[RelPath]SourceStateEntry{
-					"script": &SourceStateFile{
+					NewRelPath("script"): &SourceStateFile{
+						origin:        "run_script",
 						sourceRelPath: NewSourceRelPath("run_script"),
 						Attr: FileAttr{
 							TargetName: "script",
@@ -929,7 +1000,7 @@ func TestSourceStateRead(t *testing.T) {
 						},
 						lazyContents: newLazyContents([]byte("# contents of script\n")),
 						targetStateEntry: &TargetStateScript{
-							name:         "script",
+							name:         NewRelPath("script"),
 							lazyContents: newLazyContents([]byte("# contents of script\n")),
 						},
 					},
@@ -945,7 +1016,8 @@ func TestSourceStateRead(t *testing.T) {
 			},
 			expectedSourceState: NewSourceState(
 				withEntries(map[RelPath]SourceStateEntry{
-					".symlink": &SourceStateFile{
+					NewRelPath(".symlink"): &SourceStateFile{
+						origin:        "symlink_dot_symlink",
 						sourceRelPath: NewSourceRelPath("symlink_dot_symlink"),
 						Attr: FileAttr{
 							TargetName: ".symlink",
@@ -970,7 +1042,8 @@ func TestSourceStateRead(t *testing.T) {
 			},
 			expectedSourceState: NewSourceState(
 				withEntries(map[RelPath]SourceStateEntry{
-					"dir": &SourceStateDir{
+					NewRelPath("dir"): &SourceStateDir{
+						origin:        "dir",
 						sourceRelPath: NewSourceRelDirPath("dir"),
 						Attr: DirAttr{
 							TargetName: "dir",
@@ -979,7 +1052,8 @@ func TestSourceStateRead(t *testing.T) {
 							perm: 0o777 &^ chezmoitest.Umask,
 						},
 					},
-					"dir/file": &SourceStateFile{
+					NewRelPath("dir/file"): &SourceStateFile{
+						origin:        "dir/file",
 						sourceRelPath: NewSourceRelPath("dir/file"),
 						Attr: FileAttr{
 							TargetName: "file",
@@ -1042,7 +1116,8 @@ func TestSourceStateRead(t *testing.T) {
 			},
 			expectedSourceState: NewSourceState(
 				withEntries(map[RelPath]SourceStateEntry{
-					"dir": &SourceStateDir{
+					NewRelPath("dir"): &SourceStateDir{
+						origin:        "exact_dir",
 						sourceRelPath: NewSourceRelDirPath("exact_dir"),
 						Attr: DirAttr{
 							TargetName: "dir",
@@ -1052,7 +1127,8 @@ func TestSourceStateRead(t *testing.T) {
 							perm: 0o777 &^ chezmoitest.Umask,
 						},
 					},
-					"dir/file1": &SourceStateFile{
+					NewRelPath("dir/file1"): &SourceStateFile{
+						origin:        "exact_dir/file1",
 						sourceRelPath: NewSourceRelPath("exact_dir/file1"),
 						Attr: FileAttr{
 							TargetName: "file1",
@@ -1064,8 +1140,8 @@ func TestSourceStateRead(t *testing.T) {
 							lazyContents: newLazyContents([]byte("# contents of dir/file1\n")),
 						},
 					},
-					"dir/file2": &SourceStateRemove{
-						targetRelPath: "dir/file2",
+					NewRelPath("dir/file2"): &SourceStateRemove{
+						targetRelPath: NewRelPath("dir/file2"),
 					},
 				}),
 				withIgnore(
@@ -1085,8 +1161,8 @@ func TestSourceStateRead(t *testing.T) {
 			},
 			expectedSourceState: NewSourceState(
 				withEntries(map[RelPath]SourceStateEntry{
-					"file": &SourceStateRemove{
-						targetRelPath: "file",
+					NewRelPath("file"): &SourceStateRemove{
+						targetRelPath: NewRelPath("file"),
 					},
 				}),
 			),
@@ -1105,8 +1181,8 @@ func TestSourceStateRead(t *testing.T) {
 			},
 			expectedSourceState: NewSourceState(
 				withEntries(map[RelPath]SourceStateEntry{
-					"file1": &SourceStateRemove{
-						targetRelPath: "file1",
+					NewRelPath("file1"): &SourceStateRemove{
+						targetRelPath: NewRelPath("file1"),
 					},
 				}),
 				withIgnore(
@@ -1162,7 +1238,8 @@ func TestSourceStateRead(t *testing.T) {
 			},
 			expectedSourceState: NewSourceState(
 				withEntries(map[RelPath]SourceStateEntry{
-					"dir": &SourceStateDir{
+					NewRelPath("dir"): &SourceStateDir{
+						origin:        "dir",
 						sourceRelPath: NewSourceRelDirPath("dir"),
 						Attr: DirAttr{
 							TargetName: "dir",
@@ -1245,7 +1322,7 @@ func TestSourceStateReadExternal(t *testing.T) {
 		expectedExternals map[RelPath]External
 	}{
 		{
-			name: "external",
+			name: "external_yaml",
 			root: map[string]interface{}{
 				"/home/user/.local/share/chezmoi": map[string]interface{}{
 					".chezmoiexternal.yaml": chezmoitest.JoinLines(
@@ -1256,7 +1333,25 @@ func TestSourceStateReadExternal(t *testing.T) {
 				},
 			},
 			expectedExternals: map[RelPath]External{
-				"file": {
+				NewRelPath("file"): {
+					Type: "file",
+					URL:  httpServer.URL + "/file",
+				},
+			},
+		},
+		{
+			name: "external_toml",
+			root: map[string]interface{}{
+				"/home/user/.local/share/chezmoi": map[string]interface{}{
+					".chezmoiexternal.toml": chezmoitest.JoinLines(
+						`[file]`,
+						`    type = "file"`,
+						`    url = "`+httpServer.URL+`/file"`,
+					),
+				},
+			},
+			expectedExternals: map[RelPath]External{
+				NewRelPath("file"): {
 					Type: "file",
 					URL:  httpServer.URL + "/file",
 				},
@@ -1274,7 +1369,7 @@ func TestSourceStateReadExternal(t *testing.T) {
 				},
 			},
 			expectedExternals: map[RelPath]External{
-				".dir/file": {
+				NewRelPath(".dir/file"): {
 					Type: "file",
 					URL:  httpServer.URL + "/file",
 				},
@@ -1344,7 +1439,7 @@ func TestSourceStateReadExternalCache(t *testing.T) {
 			)
 			require.NoError(t, s.Read(ctx, readOptions))
 			assert.Equal(t, map[RelPath]External{
-				".dir": {
+				NewRelPath(".dir"): {
 					Type:          "archive",
 					URL:           httpServer.URL + "/archive.tar",
 					RefreshPeriod: 1 * time.Minute,
@@ -1392,15 +1487,15 @@ func TestSourceStateTargetRelPaths(t *testing.T) {
 				},
 			},
 			expectedTargetRelPaths: []RelPath{
-				"1before",
-				"2before",
-				"3before",
-				"1",
-				"2",
-				"3",
-				"1after",
-				"2after",
-				"3after",
+				NewRelPath("1before"),
+				NewRelPath("2before"),
+				NewRelPath("3before"),
+				NewRelPath("1"),
+				NewRelPath("2"),
+				NewRelPath("3"),
+				NewRelPath("1after"),
+				NewRelPath("2after"),
+				NewRelPath("3after"),
 			},
 		},
 	} {
@@ -1437,19 +1532,25 @@ func (s *SourceState) applyAll(targetSystem, destSystem System, persistentState
 // without error.
 func requireEvaluateAll(t *testing.T, s *SourceState, destSystem System) {
 	t.Helper()
-	for _, targetRelPath := range s.TargetRelPaths() {
-		sourceStateEntry := s.entries[targetRelPath]
-		require.NoError(t, sourceStateEntry.Evaluate())
+	require.NoError(t, s.root.ForEach(EmptyRelPath, func(targetRelPath RelPath, sourceStateEntry SourceStateEntry) error {
+		if err := sourceStateEntry.Evaluate(); err != nil {
+			return err
+		}
 		destAbsPath := s.destDirAbsPath.Join(targetRelPath)
 		targetStateEntry, err := sourceStateEntry.TargetStateEntry(destSystem, destAbsPath)
-		require.NoError(t, err)
-		require.NoError(t, targetStateEntry.Evaluate())
-	}
+		if err != nil {
+			return err
+		}
+		return targetStateEntry.Evaluate()
+	}))
 }
 
 func withEntries(sourceEntries map[RelPath]SourceStateEntry) SourceStateOption {
 	return func(s *SourceState) {
-		s.entries = sourceEntries
+		s.root = sourceStateEntryTreeNode{}
+		for targetRelPath, sourceStateEntry := range sourceEntries {
+			s.root.Set(targetRelPath, sourceStateEntry)
+		}
 	}
 }
 
