@@ -27,6 +27,7 @@ import (
 	"github.com/coreos/go-semver/semver"
 	gogit "github.com/go-git/go-git/v5"
 	"github.com/go-git/go-git/v5/plumbing/format/diff"
+	"github.com/google/gops/agent"
 	"github.com/mitchellh/mapstructure"
 	"github.com/rs/zerolog"
 	"github.com/rs/zerolog/log"
@@ -84,6 +85,7 @@ type Config struct {
 	debug            bool
 	dryRun           bool
 	force            bool
+	gops             bool
 	homeDir          string
 	keepGoing        bool
 	noPager          bool
@@ -178,8 +180,10 @@ type configState struct {
 }
 
 var (
-	persistentStateFilename = chezmoi.RelPath("chezmoistate.boltdb")
-	configStateKey          = []byte("configState")
+	chezmoiRelPath             = chezmoi.NewRelPath("chezmoi")
+	persistentStateFileRelPath = chezmoi.NewRelPath("chezmoistate.boltdb")
+
+	configStateKey = []byte("configState")
 
 	defaultAgeEncryptionConfig = chezmoi.AgeEncryption{
 		Command: "age",
@@ -224,7 +228,7 @@ func newConfig(options ...configOption) (*Config, error) {
 
 	c := &Config{
 		// Global configuration, settable in the config file.
-		CacheDirAbsPath: chezmoi.NewAbsPath(bds.CacheHome).Join("chezmoi"),
+		CacheDirAbsPath: chezmoi.NewAbsPath(bds.CacheHome).Join(chezmoiRelPath),
 		Color: autoBool{
 			auto: true,
 		},
@@ -363,6 +367,10 @@ func newConfig(options ...configOption) (*Config, error) {
 			include:   chezmoi.NewEntryTypeSet(chezmoi.EntryTypesAll),
 			recursive: true,
 		},
+		upgrade: upgradeCmdConfig{
+			owner: "twpayne",
+			repo:  "chezmoi",
+		},
 		verify: verifyCmdConfig{
 			exclude:   chezmoi.NewEntryTypeSet(chezmoi.EntryTypesNone),
 			include:   chezmoi.NewEntryTypeSet(chezmoi.EntryTypesAll &^ chezmoi.EntryTypeScripts),
@@ -476,7 +484,7 @@ func (c *Config) applyArgs(ctx context.Context, targetSystem chezmoi.System, tar
 	if err != nil {
 		return err
 	}
-	if configTemplateRelPath != "" {
+	if configTemplateRelPath != chezmoi.EmptyRelPath {
 		currentConfigTemplateContentsSHA256 = chezmoi.SHA256Sum(configTemplateContents)
 	}
 	var previousConfigTemplateContentsSHA256 []byte
@@ -493,7 +501,7 @@ func (c *Config) applyArgs(ctx context.Context, targetSystem chezmoi.System, tar
 		bytes.Equal(currentConfigTemplateContentsSHA256, previousConfigTemplateContentsSHA256)
 	if !configTemplateContentsUnchanged {
 		if c.force {
-			if configTemplateRelPath == "" {
+			if configTemplateRelPath == chezmoi.EmptyRelPath {
 				if err := c.persistentState.Delete(chezmoi.ConfigStateBucket, configStateKey); err != nil {
 					return err
 				}
@@ -518,7 +526,7 @@ func (c *Config) applyArgs(ctx context.Context, targetSystem chezmoi.System, tar
 		return err
 	}
 
-	var targetRelPaths []chezmoi.RelPath
+	var targetRelPaths chezmoi.RelPaths
 	switch {
 	case len(args) == 0:
 		targetRelPaths = sourceState.TargetRelPaths()
@@ -574,6 +582,7 @@ func (c *Config) close() error {
 		err = multierr.Append(err, err2)
 	}
 	pprof.StopCPUProfile()
+	agent.Close()
 	return err
 }
 
@@ -610,7 +619,7 @@ func (c *Config) createAndReloadConfigFile() error {
 		return err
 	}
 	var configFileContents []byte
-	if configTemplateRelPath == "" {
+	if configTemplateRelPath == chezmoi.EmptyRelPath {
 		if err := c.persistentState.Delete(chezmoi.ConfigStateBucket, configStateKey); err != nil {
 			return err
 		}
@@ -633,7 +642,7 @@ func (c *Config) createAndReloadConfigFile() error {
 		// Write the config.
 		configPath := c.init.configPath
 		if c.init.configPath.Empty() {
-			configPath = chezmoi.NewAbsPath(c.bds.ConfigHome).Join("chezmoi").Join(configTemplateRelPath)
+			configPath = chezmoi.NewAbsPath(c.bds.ConfigHome).Join(chezmoiRelPath, configTemplateRelPath)
 		}
 		if err := chezmoi.MkdirAll(c.baseSystem, configPath.Dir(), 0o777); err != nil {
 			return err
@@ -654,7 +663,7 @@ func (c *Config) createAndReloadConfigFile() error {
 	}
 
 	// Reload config if it was created.
-	if configTemplateRelPath != "" {
+	if configTemplateRelPath != chezmoi.EmptyRelPath {
 		viper.SetConfigType(ext)
 		if err := viper.ReadConfig(bytes.NewBuffer(configFileContents)); err != nil {
 			return err
@@ -680,7 +689,7 @@ func (c *Config) createConfigFile(filename chezmoi.RelPath, data []byte) ([]byte
 		"writeToStdout": c.writeToStdout,
 	})
 
-	t, err := template.New(string(filename)).Funcs(funcMap).Parse(string(data))
+	t, err := template.New(filename.String()).Funcs(funcMap).Parse(string(data))
 	if err != nil {
 		return nil, err
 	}
@@ -706,7 +715,7 @@ func (c *Config) defaultConfigFile(fileSystem vfs.Stater, bds *xdg.BaseDirectory
 			return chezmoi.EmptyAbsPath, err
 		}
 		for _, extension := range viper.SupportedExts {
-			configFileAbsPath := configDirAbsPath.JoinStr("chezmoi", "chezmoi."+extension)
+			configFileAbsPath := configDirAbsPath.JoinString("chezmoi", "chezmoi."+extension)
 			if _, err := fileSystem.Stat(configFileAbsPath.String()); err == nil {
 				return configFileAbsPath, nil
 			}
@@ -717,7 +726,7 @@ func (c *Config) defaultConfigFile(fileSystem vfs.Stater, bds *xdg.BaseDirectory
 	if err != nil {
 		return chezmoi.EmptyAbsPath, err
 	}
-	return configHomeAbsPath.Join("chezmoi", "chezmoi.toml"), nil
+	return configHomeAbsPath.JoinString("chezmoi", "chezmoi.toml"), nil
 }
 
 // defaultPreApplyFunc is the default pre-apply function. If the target entry
@@ -783,7 +792,7 @@ func (c *Config) defaultSourceDir(fileSystem vfs.Stater, bds *xdg.BaseDirectoryS
 		if err != nil {
 			return chezmoi.EmptyAbsPath, err
 		}
-		sourceDirAbsPath := dataDirAbsPath.Join("chezmoi")
+		sourceDirAbsPath := dataDirAbsPath.Join(chezmoiRelPath)
 		if _, err := fileSystem.Stat(sourceDirAbsPath.String()); err == nil {
 			return sourceDirAbsPath, nil
 		}
@@ -793,7 +802,7 @@ func (c *Config) defaultSourceDir(fileSystem vfs.Stater, bds *xdg.BaseDirectoryS
 	if err != nil {
 		return chezmoi.EmptyAbsPath, err
 	}
-	return dataHomeAbsPath.Join("chezmoi"), nil
+	return dataHomeAbsPath.Join(chezmoiRelPath), nil
 }
 
 // defaultTemplateData returns the default template data.
@@ -880,6 +889,7 @@ func (c *Config) defaultTemplateData() map[string]interface{} {
 	return map[string]interface{}{
 		"chezmoi": map[string]interface{}{
 			"arch":         runtime.GOARCH,
+			"args":         os.Args,
 			"fqdnHostname": fqdnHostname,
 			"group":        group,
 			"homeDir":      c.homeDir,
@@ -1064,17 +1074,17 @@ func (c *Config) filterInput(args []string, f func([]byte) ([]byte, error)) erro
 // format, and contents of the first one that it finds.
 func (c *Config) findFirstConfigTemplate() (chezmoi.RelPath, string, []byte, error) {
 	for _, ext := range viper.SupportedExts {
-		filename := chezmoi.RelPath(chezmoi.Prefix + "." + ext + chezmoi.TemplateSuffix)
+		filename := chezmoi.NewRelPath(chezmoi.Prefix + "." + ext + chezmoi.TemplateSuffix)
 		contents, err := c.baseSystem.ReadFile(c.SourceDirAbsPath.Join(filename))
 		switch {
 		case errors.Is(err, fs.ErrNotExist):
 			continue
 		case err != nil:
-			return "", "", nil, err
+			return chezmoi.EmptyRelPath, "", nil, err
 		}
-		return chezmoi.RelPath("chezmoi." + ext), ext, contents, nil
+		return chezmoi.NewRelPath("chezmoi." + ext), ext, contents, nil
 	}
-	return "", "", nil, nil
+	return chezmoi.EmptyRelPath, "", nil, nil
 }
 
 // gitAutoAdd adds all changes to the git index and returns the new git status.
@@ -1191,6 +1201,7 @@ func (c *Config) newRootCmd() (*cobra.Command, error) {
 	persistentFlags.BoolVar(&c.debug, "debug", c.debug, "Include debug information in output")
 	persistentFlags.BoolVarP(&c.dryRun, "dry-run", "n", c.dryRun, "Do not make any modifications to the destination directory")
 	persistentFlags.BoolVar(&c.force, "force", c.force, "Make all changes without prompting")
+	persistentFlags.BoolVar(&c.gops, "gops", c.gops, "Enable gops agent")
 	persistentFlags.BoolVarP(&c.keepGoing, "keep-going", "k", c.keepGoing, "Keep going as far as possible after an error")
 	persistentFlags.BoolVar(&c.noPager, "no-pager", c.noPager, "Do not use the pager")
 	persistentFlags.BoolVar(&c.noTTY, "no-tty", c.noTTY, "Do not attempt to get a TTY for reading passwords")
@@ -1204,6 +1215,7 @@ func (c *Config) newRootCmd() (*cobra.Command, error) {
 		rootCmd.MarkPersistentFlagFilename("cpu-profile"),
 		persistentFlags.MarkHidden("cpu-profile"),
 		rootCmd.MarkPersistentFlagDirname("destination"),
+		persistentFlags.MarkHidden("gops"),
 		rootCmd.MarkPersistentFlagFilename("output"),
 		persistentFlags.MarkHidden("safe"),
 		rootCmd.MarkPersistentFlagDirname("source"),
@@ -1372,6 +1384,13 @@ func (c *Config) persistentPreRunRootE(cmd *cobra.Command, args []string) error
 		}
 	}
 
+	// Enable gops if configured.
+	if c.gops {
+		if err := agent.Listen(agent.Options{}); err != nil {
+			return err
+		}
+	}
+
 	// Read the config file.
 	if err := c.readConfig(); err != nil {
 		if !boolAnnotation(cmd, doesNotRequireValidConfig) {
@@ -1550,7 +1569,7 @@ func (c *Config) persistentPreRunRootE(cmd *cobra.Command, args []string) error
 		workingTreeAbsPath := c.SourceDirAbsPath
 	FOR:
 		for {
-			if info, err := c.baseSystem.Stat(workingTreeAbsPath.Join(gogit.GitDirName)); err == nil && info.IsDir() {
+			if info, err := c.baseSystem.Stat(workingTreeAbsPath.JoinString(gogit.GitDirName)); err == nil && info.IsDir() {
 				c.WorkingTreeAbsPath = workingTreeAbsPath
 				break FOR
 			}
@@ -1581,14 +1600,14 @@ func (c *Config) persistentPreRunRootE(cmd *cobra.Command, args []string) error
 // none are found.
 func (c *Config) persistentStateFile() (chezmoi.AbsPath, error) {
 	if !c.configFileAbsPath.Empty() {
-		return c.configFileAbsPath.Dir().Join(persistentStateFilename), nil
+		return c.configFileAbsPath.Dir().Join(persistentStateFileRelPath), nil
 	}
 	for _, configDir := range c.bds.ConfigDirs {
 		configDirAbsPath, err := chezmoi.NewAbsPathFromExtPath(configDir, c.homeDirAbsPath)
 		if err != nil {
 			return chezmoi.EmptyAbsPath, err
 		}
-		persistentStateFile := configDirAbsPath.Join("chezmoi", persistentStateFilename)
+		persistentStateFile := configDirAbsPath.Join(chezmoiRelPath, persistentStateFileRelPath)
 		if _, err := os.Stat(persistentStateFile.String()); err == nil {
 			return persistentStateFile, nil
 		}
@@ -1597,7 +1616,7 @@ func (c *Config) persistentStateFile() (chezmoi.AbsPath, error) {
 	if err != nil {
 		return chezmoi.EmptyAbsPath, err
 	}
-	return defaultConfigFileAbsPath.Dir().Join(persistentStateFilename), nil
+	return defaultConfigFileAbsPath.Dir().Join(persistentStateFileRelPath), nil
 }
 
 // promptChoice prompts the user for one of choices until a valid choice is made.
@@ -1702,9 +1721,9 @@ type targetRelPathsOptions struct {
 }
 
 // targetRelPaths returns the target relative paths for each target path in
-// args.
-func (c *Config) targetRelPaths(sourceState *chezmoi.SourceState, args []string, options targetRelPathsOptions) ([]chezmoi.RelPath, error) {
-	targetRelPaths := make([]chezmoi.RelPath, 0, len(args))
+// args. The returned paths are sorted and de-duplicated.
+func (c *Config) targetRelPaths(sourceState *chezmoi.SourceState, args []string, options targetRelPathsOptions) (chezmoi.RelPaths, error) {
+	targetRelPaths := make(chezmoi.RelPaths, 0, len(args))
 	for _, arg := range args {
 		argAbsPath, err := chezmoi.NewAbsPathFromExtPath(arg, c.homeDirAbsPath)
 		if err != nil {
@@ -1718,7 +1737,7 @@ func (c *Config) targetRelPaths(sourceState *chezmoi.SourceState, args []string,
 			return nil, err
 		}
 		if options.mustBeInSourceState {
-			if _, ok := sourceState.Entry(targetRelPath); !ok {
+			if !sourceState.Contains(targetRelPath) {
 				return nil, fmt.Errorf("%s: not in source state", arg)
 			}
 		}
@@ -1740,9 +1759,7 @@ func (c *Config) targetRelPaths(sourceState *chezmoi.SourceState, args []string,
 	}
 
 	// Sort and de-duplicate targetRelPaths in place.
-	sort.Slice(targetRelPaths, func(i, j int) bool {
-		return targetRelPaths[i] < targetRelPaths[j]
-	})
+	sort.Sort(targetRelPaths)
 	n := 1
 	for i := 1; i < len(targetRelPaths); i++ {
 		if targetRelPaths[i] != targetRelPaths[i-1] {
@@ -1758,10 +1775,11 @@ func (c *Config) targetRelPaths(sourceState *chezmoi.SourceState, args []string,
 func (c *Config) targetRelPathsBySourcePath(sourceState *chezmoi.SourceState, args []string) ([]chezmoi.RelPath, error) {
 	targetRelPaths := make([]chezmoi.RelPath, 0, len(args))
 	targetRelPathsBySourceRelPath := make(map[chezmoi.RelPath]chezmoi.RelPath)
-	for targetRelPath, sourceStateEntry := range sourceState.Entries() {
+	_ = sourceState.ForEach(func(targetRelPath chezmoi.RelPath, sourceStateEntry chezmoi.SourceStateEntry) error {
 		sourceRelPath := sourceStateEntry.SourceRelPath().RelPath()
 		targetRelPathsBySourceRelPath[sourceRelPath] = targetRelPath
-	}
+		return nil
+	})
 	for _, arg := range args {
 		argAbsPath, err := chezmoi.NewAbsPathFromExtPath(arg, c.homeDirAbsPath)
 		if err != nil {
