@@ -11,7 +11,7 @@ import (
 	"github.com/twpayne/chezmoi/v2/internal/chezmoi"
 )
 
-func (c *umaskCheck) Run(system chezmoi.System) (checkResult, string) {
+func (c *umaskCheck) Run(system chezmoi.System, homeDirAbsPath chezmoi.AbsPath) (checkResult, string) {
 	umask := unix.Umask(0)
 	unix.Umask(umask)
 	result := checkResultOK
