@@ -7,7 +7,11 @@ import (
 	"github.com/spf13/cobra"
 )
 
-type upgradeCmdConfig struct{}
+type upgradeCmdConfig struct {
+	method string
+	owner  string
+	repo   string
+}
 
 func (c *Config) newUpgradeCmd() *cobra.Command {
 	return nil
