@@ -33,10 +33,11 @@ func (c *Config) runUnmanagedCmd(cmd *cobra.Command, args []string, sourceState
 			return nil
 		}
 		targeRelPath := destAbsPath.MustTrimDirPrefix(c.DestDirAbsPath)
-		_, managed := sourceState.Entry(targeRelPath)
+		managed := sourceState.Contains(targeRelPath)
 		ignored := sourceState.Ignore(targeRelPath)
 		if !managed && !ignored {
-			builder.WriteString(string(targeRelPath) + "\n")
+			builder.WriteString(targeRelPath.String())
+			builder.WriteByte('\n')
 		}
 		if info.IsDir() && (!managed || ignored) {
 			return vfs.SkipDir
