@@ -45,14 +45,14 @@ func lintFile(filename string) error {
 	for i, line := range lines {
 		switch {
 		case crlfLineEndingRx.Match(line):
-			err = multierr.Append(err, fmt.Errorf("%s:%d: CRLF line ending", filename, i+1))
+			err = multierr.Append(err, fmt.Errorf("::error file=%s,line=%d::CRLF line ending", filename, i+1))
 		case trailingWhitespaceRx.Match(line):
-			err = multierr.Append(err, fmt.Errorf("%s:%d: trailing whitespace", filename, i+1))
+			err = multierr.Append(err, fmt.Errorf("::error file=%s,line=%d::trailing whitespace", filename, i+1))
 		}
 	}
 
 	if len(data) > 0 && len(lines[len(lines)-1]) != 0 {
-		err = multierr.Append(err, fmt.Errorf("%s: no newline at end of file", filename))
+		err = multierr.Append(err, fmt.Errorf("::error file=%s,line=%d::no newline at end of file", filename, len(lines)+1))
 	}
 
 	return err
